import { createRoot } from "react-dom/client";
import { HelmetProvider } from "react-helmet-async";
import App from "./App";
import "./index.css";

// Create a helmet context to use for server-like behavior
const helmetContext = {};

// Function to update document head manually with Helmet data - optimized
function updateDocumentHead(helmetContext: any) {
  if (!helmetContext.helmet) return;
  
  // Use requestAnimationFrame to batch DOM updates and avoid layout thrashing
  requestAnimationFrame(() => {
    const { helmet } = helmetContext;
    
    // Update title
    if (helmet.title.toString()) {
      document.title = helmet.title.toString();
    }
    
    // Create a map of current meta tags for faster lookups instead of multiple DOM queries
    const currentMetaTags = new Map();
    const metaTags = document.head.querySelectorAll('meta[data-react-helmet="true"]');
    
    metaTags.forEach(tag => {
      const name = tag.getAttribute('name');
      const property = tag.getAttribute('property');
      const key = name || property;
      if (key) currentMetaTags.set(key, tag);
    });
    
    // Parse and process new meta tags
    const metaString = helmet.meta.toString();
    if (metaString) {
      const parser = new DOMParser();
      const metaFragment = parser.parseFromString(`<head>${metaString}</head>`, 'text/html');
      const newMetaTags = metaFragment.head.querySelectorAll('meta');
      
      // Create a document fragment to batch DOM insertions
      const fragment = document.createDocumentFragment();
      const processedKeys = new Set();
      
      newMetaTags.forEach(tag => {
        const name = tag.getAttribute('name');
        const property = tag.getAttribute('property');
        const key = name || property;
        
        if (key) {
          processedKeys.add(key);
          const existingTag = currentMetaTags.get(key);
          
          if (existingTag) {
            // Update attributes of existing tag
            Array.from(tag.attributes).forEach(attr => {
              if (attr.name !== 'data-react-helmet') {
                existingTag.setAttribute(attr.name, attr.value);
              }
            });
          } else {
            // Create new tag
            const newTag = document.createElement('meta');
            Array.from(tag.attributes).forEach(attr => {
              newTag.setAttribute(attr.name, attr.value);
            });
            newTag.setAttribute('data-react-helmet', 'true');
            fragment.appendChild(newTag);
          }
        }
      });
      
      // Remove meta tags that are no longer needed
      currentMetaTags.forEach((tag, key) => {
        if (!processedKeys.has(key)) {
          tag.remove();
        }
      });
      
      // Append all new tags at once
      if (fragment.childNodes.length > 0) {
        document.head.appendChild(fragment);
      }
    }
  });
}

// Avoid excessive updates by debouncing the observer
let updateTimeout: number | undefined;
const DEBOUNCE_TIME = 100; // ms

const bodyObserver = new MutationObserver(() => {
  // Debounce multiple rapid mutations 
  if (updateTimeout) window.clearTimeout(updateTimeout);
  updateTimeout = window.setTimeout(() => {
    updateDocumentHead(helmetContext);
  }, DEBOUNCE_TIME);
});

// Start observing the document body for DOM changes - only observe necessary changes
bodyObserver.observe(document.body, {
  childList: true,      // React component mounts/unmounts
  subtree: false,       // Don't observe all descendants, too expensive
  attributes: false     // Don't care about attribute changes for meta tag updating
});

// Apply initial meta tags and remove any loading indicators 
window.addEventListener('load', () => {
  // Update meta tags
  updateDocumentHead(helmetContext);
  
  // Remove skeleton placeholders when actual content is ready
  const root = document.getElementById('root');
  const loadingIndicator = document.querySelector('.loading-indicator');
  if (root && loadingIndicator && root.contains(loadingIndicator)) {
    loadingIndicator.classList.add('fade-out');
    setTimeout(() => {
      if (loadingIndicator.parentNode) {
        loadingIndicator.parentNode.removeChild(loadingIndicator);
      }
    }, 300); // Allow time for fade out animation
  }
});

// Initialize Tailwind components and custom CSS
createRoot(document.getElementById("root")!).render(
  <HelmetProvider context={helmetContext}>
    <App />
  </HelmetProvider>
);
